/// <reference path="../types.js" />

import { stringifyQuery } from './query';

const trailingSlashRE = /\/?$/;

/**
 * @param {RouteRecord} [record]
 * @param {Location} location
 * @param {Location} [redirectedFrom]
 * @param {VueRouter} router
 * @returns {Route}
 */
export function createRoute(record, location, redirectedFrom, router) {
  const stringifyQuery = router && router.options.stringifyQuery;

  /** @type {any} */
  let query = location.query || {};
  try {
    query = clone(query);
  } catch (e) {}

  /** @type {Route} */
  const route = {
    name: location.name || (record && record.name),
    meta: (record && record.meta) || {},
    path: location.path || '/',
    hash: location.hash || '',
    query,
    params: location.params || {},
    fullPath: getFullPath(location, stringifyQuery),
    matched: record ? formatMatch(record) : [],
  };

  if (redirectedFrom) {
    route.redirectedFrom = getFullPath(redirectedFrom, stringifyQuery);
  }

  return Object.freeze(route);
}

function clone(value) {
  if (Array.isArray(value)) {
    return value.map(clone);
  } else if (value && typeof value === 'object') {
    const res = {};
    for (const key in value) {
      res[key] = clone(value[key]);
    }
    return res;
  } else {
    return value;
  }
}

// the starting route that represents the initial state
export const START = createRoute(null, {
  path: '/',
});

/**
 * @param {RouteRecord} [record]
 * @returns {RouteRecord}
 */
function formatMatch(record) {
  const res = [];
  while (record) {
    res.unshift(record);
    record = record.parent;
  }
  return res;
}

/**
 *
 * @param {{ path?: string, query?: any, hash?: string }} param0
 * @param {Function} [_stringifyQuery]
 * @returns {string}
 */
function getFullPath({ path, query = {}, hash = '' }, _stringifyQuery) {
  const stringify = _stringifyQuery || stringifyQuery;
  return (path || '/') + stringify(query) + hash;
}

/**
 * @param {Route} a
 * @param {Route} [b]
 * @param {boolean} [onlyPath]
 * @returns {boolean}
 */
export function isSameRoute(a, b, onlyPath) {
  if (b === START) {
    return a === b;
  } else if (!b) {
    return false;
  } else if (a.path && b.path) {
    return (
      a.path.replace(trailingSlashRE, '') ===
        b.path.replace(trailingSlashRE, '') &&
      (onlyPath || (a.hash === b.hash && isObjectEqual(a.query, b.query)))
    );
  } else if (a.name && b.name) {
    return (
      a.name === b.name &&
      (onlyPath ||
        (a.hash === b.hash &&
          isObjectEqual(a.query, b.query) &&
          isObjectEqual(a.params, b.params)))
    );
  } else {
    return false;
  }
}

/**
 * @param {any} a
 * @param {any} b
 * @returns {boolean}
 */
function isObjectEqual(a = {}, b = {}) {
  // handle null value #1566
  if (!a || !b) return a === b;
  const aKeys = Object.keys(a).sort();
  const bKeys = Object.keys(b).sort();
  if (aKeys.length !== bKeys.length) {
    return false;
  }
  return aKeys.every((key, i) => {
    const aVal = a[key];
    const bKey = bKeys[i];
    if (bKey !== key) return false;
    const bVal = b[key];
    // query values can be null and undefined
    if (aVal == null || bVal == null) return aVal === bVal;
    // check nested equality
    if (typeof aVal === 'object' && typeof bVal === 'object') {
      return isObjectEqual(aVal, bVal);
    }
    return String(aVal) === String(bVal);
  });
}

/**
 * @param {Route} current
 * @param {Route} target
 * @returns {boolean}
 */
export function isIncludedRoute(current, target) {
  return (
    current.path
      .replace(trailingSlashRE, '/')
      .indexOf(target.path.replace(trailingSlashRE, '/')) === 0 &&
    (!target.hash || current.hash === target.hash) &&
    queryIncludes(current.query, target.query)
  );
}

/**
 * @param {Dictionary<string>} current
 * @param {Dictionary<string>} target
 * @returns {boolean}
 */
function queryIncludes(current, target) {
  for (const key in target) {
    if (!(key in current)) {
      return false;
    }
  }
  return true;
}

/**
 * @param {Route} route
 */
export function handleRouteEntered(route) {
  for (let i = 0; i < route.matched.length; i++) {
    const record = route.matched[i];
    for (const name in record.instances) {
      const instance = record.instances[name];
      const cbs = record.enteredCbs[name];
      if (!instance || !cbs) continue;
      delete record.enteredCbs[name];
      for (let i = 0; i < cbs.length; i++) {
        if (!instance._isBeingDestroyed) cbs[i](instance);
      }
    }
  }
}
